@@ -73,7 +73,7 @@ check_pid_and_port()
             grep -E '[[:space:]]+(rsync|stunnel)[[:space:]]+'"$rsync_pid" 2>/dev/null)"
         ;;
     *)
-        if ! which lsof > /dev/null; then
+        if ! command -v lsof > /dev/null; then
           wsrep_log_error "lsof tool not found in PATH! Make sure you have it installed."
           exit 2 # ENOENT
         fi
@@ -104,10 +104,10 @@ check_pid_and_port()
 is_local_ip()
 {
   local address="$1"
-  local get_addr_bin=`which ifconfig`
-  if [ -z "$get_addr_bin" ]
+  local get_addr_bin
+  if ! command -v ifconfig > /dev/null
   then
-    get_addr_bin=`which ip`
+    get_addr_bin=ip
     get_addr_bin="$get_addr_bin address show"
     # Add an slash at the end, so we don't get false positive : 172.18.0.4 matches 172.18.0.41
     # ip output format is "X.X.X.X/mask"
@@ -115,6 +115,7 @@ is_local_ip()
   else
     # Add an space at the end, so we don't get false positive : 172.18.0.4 matches 172.18.0.41
     # ifconfig output format is "X.X.X.X "
+    get_addr_bin=ifconfig
     address="$address "
   fi
 
@@ -146,6 +147,7 @@ if ! [ -z $WSREP_SST_OPT_BINLOG_INDEX ]
 then
     BINLOG_INDEX_DIRNAME=$(dirname $WSREP_SST_OPT_BINLOG_INDEX)
     BINLOG_INDEX_FILENAME=$(basename $WSREP_SST_OPT_BINLOG_INDEX)
+    BINLOG_INDEX_FILENAME=${BINLOG_INDEX_FILENAME%.index}.index
 fi
 
 WSREP_LOG_DIR=${WSREP_LOG_DIR:-""}
@@ -277,11 +279,12 @@ EOF
             OLD_PWD="$(pwd)"
             cd $BINLOG_DIRNAME
 
-            if ! [ -z $WSREP_SST_OPT_BINLOG_INDEX ]
-               binlog_files_full=$(tail -n $BINLOG_N_FILES ${BINLOG_FILENAME}.index)
+            if [ -z $WSREP_SST_OPT_BINLOG_INDEX ]
             then
-               cd $BINLOG_INDEX_DIRNAME
-               binlog_files_full=$(tail -n $BINLOG_N_FILES ${BINLOG_INDEX_FILENAME}.index)
+                binlog_files_full=$(tail -n $BINLOG_N_FILES ${BINLOG_FILENAME}.index)
+            else
+                cd $BINLOG_INDEX_DIRNAME
+                binlog_files_full=$(tail -n $BINLOG_N_FILES ${BINLOG_INDEX_FILENAME})
             fi
 
             cd $BINLOG_DIRNAME
@@ -337,11 +340,11 @@ EOF
             exit 255 # unknown error
         fi
 
-        # second, we transfer InnoDB log files
+        # second, we transfer InnoDB and Aria log files
         rsync ${STUNNEL:+--rsh="$STUNNEL"} \
               --owner --group --perms --links --specials \
               --ignore-times --inplace --dirs --delete --quiet \
-              $WHOLE_FILE_OPT -f '+ /ib_logfile[0-9]*' -f '- **' "$WSREP_LOG_DIR/" \
+              $WHOLE_FILE_OPT -f '+ /ib_logfile[0-9]*' -f '+ /aria_log.*' -f '+ /aria_log_control' -f '- **' "$WSREP_LOG_DIR/" \
               rsync://$WSREP_SST_OPT_ADDR-log_dir >&2 || RC=$?
 
         if [ $RC -ne 0 ]; then
@@ -362,7 +365,7 @@ EOF
              rsync ${STUNNEL:+--rsh="$STUNNEL"} \
              --owner --group --perms --links --specials \
              --ignore-times --inplace --recursive --delete --quiet \
-             $WHOLE_FILE_OPT --exclude '*/ib_logfile*' "$WSREP_SST_OPT_DATA"/{}/ \
+             $WHOLE_FILE_OPT --exclude '*/ib_logfile*' --exclude "*/aria_log.*" --exclude "*/aria_log_control" "$WSREP_SST_OPT_DATA"/{}/ \
              rsync://$WSREP_SST_OPT_ADDR/{} >&2 || RC=$?
 
         cd "$OLD_PWD"
@@ -398,6 +401,14 @@ then
     MODULE="rsync_sst"
 
     RSYNC_PID="$WSREP_SST_OPT_DATA/$MODULE.pid"
+    # give some time for lingering rsync from previous SST to complete
+    check_round=0
+    while check_pid $RSYNC_PID && [ $check_round -lt 10 ]
+    do
+        wsrep_log_info "lingering rsync daemon found at startup, waiting for it to exit"
+        check_round=$(( check_round + 1 ))
+        sleep 1
+    done
 
     if check_pid $RSYNC_PID
     then
@@ -519,15 +530,16 @@ EOF
             # Clean up old binlog files first
             rm -f ${BINLOG_FILENAME}.*
             wsrep_log_info "Extracting binlog files:"
-            tar -xvf $BINLOG_TAR_FILE >&2
-            for ii in $(ls -1 ${BINLOG_FILENAME}.*)
-            do
-                if ! [ -z $WSREP_SST_OPT_BINLOG_INDEX ]
-                  echo ${BINLOG_DIRNAME}/${ii} >> ${BINLOG_FILENAME}.index
-		then
-                  echo ${BINLOG_DIRNAME}/${ii} >> ${BINLOG_INDEX_DIRNAME}/${BINLOG_INDEX_FILENAME}.index
+            tar -xvf $BINLOG_TAR_FILE >> _binlog_tmp_files_$!
+            while read bin_file; do
+                if [ -z $WSREP_SST_OPT_BINLOG_INDEX ]
+                then
+                    echo ${BINLOG_DIRNAME}/${bin_file} >> ${BINLOG_FILENAME}.index
+                else
+                    echo ${BINLOG_DIRNAME}/${bin_file} >> ${BINLOG_INDEX_DIRNAME}/${BINLOG_INDEX_FILENAME}
                 fi
-            done
+            done < _binlog_tmp_files_$!
+            rm -f _binlog_tmp_files_$!
         fi
         cd "$OLD_PWD"
 
