@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2019, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -39,9 +39,11 @@ Created 3/26/1996 Heikki Tuuri
 #include "row0row.h"
 #include "row0mysql.h"
 
-/** The search tuple corresponding to TRX_UNDO_INSERT_METADATA */
+/** The search tuple corresponding to TRX_UNDO_INSERT_METADATA. */
 const dtuple_t trx_undo_metadata = {
-	REC_INFO_METADATA, 0, 0,
+	/* This also works for REC_INFO_METADATA_ALTER, because the
+	delete-mark (REC_INFO_DELETED_FLAG) is ignored when searching. */
+	REC_INFO_METADATA_ADD, 0, 0,
 	NULL, 0, NULL
 #ifdef UNIV_DEBUG
 	, DATA_TUPLE_MAGIC_N
@@ -141,8 +143,8 @@ trx_undo_parse_add_undo_rec(
 }
 
 /** Calculate the free space left for extending an undo log record.
-@param block   undo log page
-@param ptr     current end of the undo page
+@param undo_block    undo log page
+@param ptr           current end of the undo page
 @return bytes left */
 static ulint trx_undo_left(const buf_block_t *undo_block, const byte *ptr)
 {
@@ -176,9 +178,6 @@ trx_undo_page_set_next_prev_and_add(
 					that points to the next free
 					offset value within undo_page.*/
 
-	ut_ad(ptr > undo_block->frame);
-	ut_ad(ptr < undo_block->frame + srv_page_size);
-
 	if (UNIV_UNLIKELY(trx_undo_left(undo_block, ptr) < 2)) {
 		return(0);
 	}
@@ -231,21 +230,35 @@ trx_undo_log_v_idx(
 {
 	ut_ad(pos < table->n_v_def);
 	dict_v_col_t*	vcol = dict_table_get_nth_v_col(table, pos);
-
-	ulint		n_idx = vcol->v_indexes->size();
 	byte*		old_ptr;
 
-	ut_ad(n_idx > 0);
+	ut_ad(!vcol->v_indexes.empty());
 
-	/* Size to reserve, max 5 bytes for each index id and position, plus
-	5 bytes for num of indexes, 2 bytes for write total length.
-	1 byte for undo log record format version marker */
-	ulint		size = n_idx * (5 + 5) + 5 + 2 + (first_v_col ? 1 : 0);
+	ulint		size = first_v_col ? 1 + 2 : 2;
+	const ulint	avail = trx_undo_left(undo_block, ptr);
 
-	if (trx_undo_left(undo_block, ptr) < size) {
+	/* The mach_write_compressed(ptr, flen) in
+	trx_undo_page_report_modify() will consume additional 1 to 5 bytes. */
+	if (avail < size + 5) {
 		return(NULL);
 	}
 
+	ulint n_idx = 0;
+	for (const auto& v_index : vcol->v_indexes) {
+		n_idx++;
+		/* FIXME: index->id is 64 bits! */
+		size += mach_get_compressed_size(uint32_t(v_index.index->id));
+		size += mach_get_compressed_size(v_index.nth_field);
+	}
+
+	size += mach_get_compressed_size(n_idx);
+
+	if (avail < size + 5) {
+		return(NULL);
+	}
+
+	ut_d(const byte* orig_ptr = ptr);
+
 	if (first_v_col) {
 		/* write the version marker */
 		mach_write_to_1(ptr, VIRTUAL_COL_UNDO_FORMAT_1);
@@ -259,18 +272,16 @@ trx_undo_log_v_idx(
 
 	ptr += mach_write_compressed(ptr, n_idx);
 
-	dict_v_idx_list::iterator       it;
-
-	for (it = vcol->v_indexes->begin();
-	     it != vcol->v_indexes->end(); ++it) {
-		dict_v_idx_t	v_index = *it;
-
+	for (const auto& v_index : vcol->v_indexes) {
 		ptr += mach_write_compressed(
-			ptr, static_cast<ulint>(v_index.index->id));
+			/* FIXME: index->id is 64 bits! */
+			ptr, uint32_t(v_index.index->id));
 
 		ptr += mach_write_compressed(ptr, v_index.nth_field);
 	}
 
+	ut_ad(orig_ptr + size == ptr);
+
 	mach_write_to_2(old_ptr, ulint(ptr - old_ptr));
 
 	return(ptr);
@@ -488,8 +499,6 @@ trx_undo_page_report_insert(
 				      + undo_block->frame);
 	ptr = undo_block->frame + first_free;
 
-	ut_ad(first_free <= srv_page_size);
-
 	if (trx_undo_left(undo_block, ptr) < 2 + 1 + 11 + 11) {
 		/* Not enough space for writing the general parameters */
 		return(0);
@@ -506,7 +515,7 @@ trx_undo_page_report_insert(
 	/* Store then the fields required to uniquely determine the record
 	to be inserted in the clustered index */
 	if (UNIV_UNLIKELY(clust_entry->info_bits != 0)) {
-		ut_ad(clust_entry->info_bits == REC_INFO_METADATA);
+		ut_ad(clust_entry->is_metadata());
 		ut_ad(index->is_instant());
 		ut_ad(undo_block->frame[first_free + 2]
 		      == TRX_UNDO_INSERT_REC);
@@ -719,7 +728,7 @@ trx_undo_rec_skip_row_ref(
 log of an update or delete marking of a clustered index record.
 @param[out]	ext_buf		buffer to hold the prefix data and BLOB pointer
 @param[in]	prefix_len	prefix size to store in the undo log
-@param[in]	page_size	page size
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in]	field		an externally stored column
 @param[in,out]	len		input: length of field; output: used length of
 ext_buf
@@ -729,13 +738,13 @@ byte*
 trx_undo_page_fetch_ext(
 	byte*			ext_buf,
 	ulint			prefix_len,
-	const page_size_t&	page_size,
+	ulint			zip_size,
 	const byte*		field,
 	ulint*			len)
 {
 	/* Fetch the BLOB. */
 	ulint	ext_len = btr_copy_externally_stored_field_prefix(
-		ext_buf, prefix_len, page_size, field, *len);
+		ext_buf, prefix_len, zip_size, field, *len);
 	/* BLOBs should always be nonempty. */
 	ut_a(ext_len);
 	/* Append the BLOB pointer to the prefix. */
@@ -753,7 +762,7 @@ available
 				size, or NULL when should not fetch a longer
 				prefix
 @param[in]	prefix_len	prefix size to store in the undo log
-@param[in]	page_size	page size
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in,out]	field		the locally stored part of the externally
 stored column
 @param[in,out]	len		length of field, in bytes
@@ -766,7 +775,7 @@ trx_undo_page_report_modify_ext(
 	byte*			ptr,
 	byte*			ext_buf,
 	ulint			prefix_len,
-	const page_size_t&	page_size,
+	ulint			zip_size,
 	const byte**		field,
 	ulint*			len,
 	spatial_status_t	spatial_status)
@@ -808,7 +817,7 @@ trx_undo_page_report_modify_ext(
 		ptr += mach_write_compressed(ptr, *len);
 
 		*field = trx_undo_page_fetch_ext(ext_buf, prefix_len,
-						 page_size, *field, len);
+						 zip_size, *field, len);
 
 		ptr += mach_write_compressed(ptr, *len + spatial_len);
 	} else {
@@ -821,7 +830,7 @@ trx_undo_page_report_modify_ext(
 
 /** Get MBR from a Geometry column stored externally
 @param[out]	mbr		MBR to fill
-@param[in]	pagesize	table pagesize
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in]	field		field contain the geometry data
 @param[in,out]	len		length of field, in bytes
 */
@@ -829,17 +838,17 @@ static
 void
 trx_undo_get_mbr_from_ext(
 /*======================*/
-	double*			mbr,
-	const page_size_t&      page_size,
-	const byte*             field,
-	ulint*			len)
+	double*		mbr,
+	ulint		zip_size,
+	const byte*	field,
+	ulint*		len)
 {
 	uchar*		dptr = NULL;
 	ulint		dlen;
 	mem_heap_t*	heap = mem_heap_create(100);
 
 	dptr = btr_copy_externally_stored_field(
-		&dlen, field, page_size, *len, heap);
+		&dlen, field, zip_size, *len, heap);
 
 	if (dlen <= GEO_DATA_HEADER_SIZE) {
 		for (uint i = 0; i < SPDIMS; ++i) {
@@ -896,8 +905,6 @@ trx_undo_page_report_modify(
 				      + undo_block->frame);
 	ptr = undo_block->frame + first_free;
 
-	ut_ad(first_free <= srv_page_size);
-
 	if (trx_undo_left(undo_block, ptr) < 50) {
 		/* NOTE: the value 50 must be big enough so that the general
 		fields written below fit on the undo log page */
@@ -923,9 +930,9 @@ trx_undo_page_report_modify(
 	/* Store first some general parameters to the undo log */
 
 	if (!update) {
-		ut_ad(!rec_get_deleted_flag(rec, dict_table_is_comp(table)));
+		ut_ad(!rec_is_delete_marked(rec, dict_table_is_comp(table)));
 		type_cmpl = TRX_UNDO_DEL_MARK_REC;
-	} else if (rec_get_deleted_flag(rec, dict_table_is_comp(table))) {
+	} else if (rec_is_delete_marked(rec, dict_table_is_comp(table))) {
 		/* In delete-marked records, DB_TRX_ID must
 		always refer to an existing update_undo log record. */
 		ut_ad(row_get_rec_trx_id(rec, index, offsets));
@@ -954,9 +961,7 @@ trx_undo_page_report_modify(
 	*ptr++ = (byte) rec_get_info_bits(rec, dict_table_is_comp(table));
 
 	/* Store the values of the system columns */
-	field = rec_get_nth_field(rec, offsets,
-				  dict_index_get_sys_col_pos(
-					  index, DATA_TRX_ID), &flen);
+	field = rec_get_nth_field(rec, offsets, index->db_trx_id(), &flen);
 	ut_ad(flen == DATA_TRX_ID_LEN);
 
 	trx_id = trx_read_trx_id(field);
@@ -970,9 +975,7 @@ trx_undo_page_report_modify(
 	}
 	ptr += mach_u64_write_compressed(ptr, trx_id);
 
-	field = rec_get_nth_field(rec, offsets,
-				  dict_index_get_sys_col_pos(
-					  index, DATA_ROLL_PTR), &flen);
+	field = rec_get_nth_field(rec, offsets, index->db_roll_ptr(), &flen);
 	ut_ad(flen == DATA_ROLL_PTR_LEN);
 	ut_ad(memcmp(field, field_ref_zero, DATA_ROLL_PTR_LEN));
 
@@ -1033,45 +1036,60 @@ trx_undo_page_report_modify(
 				on them */
 				if (upd_fld_is_virtual_col(fld)
 				    && dict_table_get_nth_v_col(
-					table, pos)->v_indexes->empty()) {
+					    table, pos)->v_indexes.empty()) {
 					n_updated--;
 				}
 			}
 		}
 
+		i = 0;
+
+		if (UNIV_UNLIKELY(update->is_alter_metadata())) {
+			ut_ad(update->n_fields >= 1);
+			ut_ad(!upd_fld_is_virtual_col(&update->fields[0]));
+			ut_ad(update->fields[0].field_no
+			      == index->first_user_field());
+			ut_ad(!dfield_is_ext(&update->fields[0].new_val));
+			ut_ad(!dfield_is_null(&update->fields[0].new_val));
+			/* The instant ADD COLUMN metadata record does not
+			contain the BLOB. Do not write anything for it. */
+			i = !rec_is_alter_metadata(rec, *index);
+			n_updated -= i;
+		}
+
 		ptr += mach_write_compressed(ptr, n_updated);
 
-		for (i = 0; i < upd_get_n_fields(update); i++) {
+		for (; i < upd_get_n_fields(update); i++) {
+			if (trx_undo_left(undo_block, ptr) < 5) {
+				return 0;
+			}
+
 			upd_field_t*	fld = upd_get_nth_field(update, i);
 
 			bool	is_virtual = upd_fld_is_virtual_col(fld);
 			ulint	max_v_log_len = 0;
 
-			ulint	pos = fld->field_no;
-
-			/* Write field number to undo log */
-			if (trx_undo_left(undo_block, ptr) < 5) {
-				return(0);
-			}
+			ulint pos = fld->field_no;
+			const dict_col_t* col = NULL;
 
 			if (is_virtual) {
 				/* Skip the non-indexed column, during
 				an online alter table */
 				if (dict_index_is_online_ddl(index)
 				    && dict_table_get_nth_v_col(
-					table, pos)->v_indexes->empty()) {
+					table, pos)->v_indexes.empty()) {
 					continue;
 				}
 
 				/* add REC_MAX_N_FIELDS to mark this
 				is a virtual col */
-				pos += REC_MAX_N_FIELDS;
-			}
+				ptr += mach_write_compressed(
+					ptr, pos + REC_MAX_N_FIELDS);
 
-			ptr += mach_write_compressed(ptr, pos);
+				if (trx_undo_left(undo_block, ptr) < 15) {
+					return 0;
+				}
 
-			/* Save the old value of field */
-			if (is_virtual) {
 				ut_ad(fld->field_no < table->n_v_def);
 
 				ptr = trx_undo_log_v_idx(undo_block, table,
@@ -1096,36 +1114,87 @@ trx_undo_page_report_modify(
 					flen = ut_min(
 						flen, max_v_log_len);
 				}
+
+				goto store_len;
+			}
+
+			if (UNIV_UNLIKELY(update->is_metadata())) {
+				ut_ad(pos >= index->first_user_field());
+				ut_ad(rec_is_metadata(rec, *index));
+
+				if (rec_is_alter_metadata(rec, *index)) {
+					ut_ad(update->is_alter_metadata());
+
+					field = rec_offs_n_fields(offsets)
+						> pos
+						&& !rec_offs_nth_default(
+							offsets, pos)
+						? rec_get_nth_field(
+							rec, offsets,
+							pos, &flen)
+						: index->instant_field_value(
+							pos - 1, &flen);
+
+					if (pos == index->first_user_field()) {
+						ut_ad(rec_offs_nth_extern(
+							offsets, pos));
+						ut_ad(flen == FIELD_REF_SIZE);
+						goto write_field;
+					}
+					col = dict_index_get_nth_col(index,
+								     pos - 1);
+				} else if (!update->is_alter_metadata()) {
+					goto get_field;
+				} else {
+					/* We are converting an ADD COLUMN
+					metadata record to an ALTER TABLE
+					metadata record, with BLOB. Subtract
+					the missing metadata BLOB field. */
+					ut_ad(pos > index->first_user_field());
+					--pos;
+					goto get_field;
+				}
 			} else {
+get_field:
+				col = dict_index_get_nth_col(index, pos);
 				field = rec_get_nth_cfield(
 					rec, index, offsets, pos, &flen);
 			}
+write_field:
+			/* Write field number to undo log */
+			ptr += mach_write_compressed(ptr, pos);
 
 			if (trx_undo_left(undo_block, ptr) < 15) {
-				return(0);
+				return 0;
 			}
 
-			if (!is_virtual && rec_offs_nth_extern(offsets, pos)) {
-				const dict_col_t*	col
-					= dict_index_get_nth_col(index, pos);
-				ulint			prefix_len
-					= dict_max_field_len_store_undo(
-						table, col);
+			if (rec_offs_n_fields(offsets) > pos
+			    && rec_offs_nth_extern(offsets, pos)) {
+				ut_ad(col || pos == index->first_user_field());
+				ut_ad(col || update->is_alter_metadata());
+				ut_ad(col
+				      || rec_is_alter_metadata(rec, *index));
+				ulint prefix_len = col
+					? dict_max_field_len_store_undo(
+						table, col)
+					: 0;
 
 				ut_ad(prefix_len + BTR_EXTERN_FIELD_REF_SIZE
 				      <= sizeof ext_buf);
 
 				ptr = trx_undo_page_report_modify_ext(
 					ptr,
-					col->ord_part
+					col
+					&& col->ord_part
 					&& !ignore_prefix
 					&& flen < REC_ANTELOPE_MAX_INDEX_COL_LEN
 					? ext_buf : NULL, prefix_len,
-					dict_table_page_size(table),
+					table->space->zip_size(),
 					&field, &flen, SPATIAL_UNKNOWN);
 
 				*type_cmpl_ptr |= TRX_UNDO_UPD_EXTERN;
 			} else {
+store_len:
 				ptr += mach_write_compressed(ptr, flen);
 			}
 
@@ -1274,6 +1343,8 @@ trx_undo_page_report_modify(
 							table, col);
 
 					ut_a(prefix_len < sizeof ext_buf);
+					const ulint zip_size
+						= table->space->zip_size();
 
 					/* If there is a spatial index on it,
 					log its MBR */
@@ -1282,9 +1353,7 @@ trx_undo_page_report_modify(
 								col->mtype));
 
 						trx_undo_get_mbr_from_ext(
-							mbr,
-							dict_table_page_size(
-								table),
+							mbr, zip_size,
 							field, &flen);
 					}
 
@@ -1293,7 +1362,7 @@ trx_undo_page_report_modify(
 						flen < REC_ANTELOPE_MAX_INDEX_COL_LEN
 						&& !ignore_prefix
 						? ext_buf : NULL, prefix_len,
-						dict_table_page_size(table),
+						zip_size,
 						&field, &flen,
 						spatial_status);
 				} else {
@@ -1490,7 +1559,6 @@ trx_undo_update_rec_get_update(
 	upd_t*		update;
 	ulint		n_fields;
 	byte*		buf;
-	ulint		i;
 	bool		first_v_col = true;
 	bool		is_undo_log = true;
 	ulint		n_skip_field = 0;
@@ -1503,7 +1571,7 @@ trx_undo_update_rec_get_update(
 		n_fields = 0;
 	}
 
-	update = upd_create(n_fields + 2, heap);
+	*upd = update = upd_create(n_fields + 2, heap);
 
 	update->info_bits = info_bits;
 
@@ -1515,9 +1583,7 @@ trx_undo_update_rec_get_update(
 
 	mach_write_to_6(buf, trx_id);
 
-	upd_field_set_field_no(upd_field,
-			       dict_index_get_sys_col_pos(index, DATA_TRX_ID),
-			       index);
+	upd_field_set_field_no(upd_field, index->db_trx_id(), index);
 	dfield_set_data(&(upd_field->new_val), buf, DATA_TRX_ID_LEN);
 
 	upd_field = upd_get_nth_field(update, n_fields + 1);
@@ -1526,25 +1592,20 @@ trx_undo_update_rec_get_update(
 
 	trx_write_roll_ptr(buf, roll_ptr);
 
-	upd_field_set_field_no(
-		upd_field, dict_index_get_sys_col_pos(index, DATA_ROLL_PTR),
-		index);
+	upd_field_set_field_no(upd_field, index->db_roll_ptr(), index);
 	dfield_set_data(&(upd_field->new_val), buf, DATA_ROLL_PTR_LEN);
 
 	/* Store then the updated ordinary columns to the update vector */
 
-	for (i = 0; i < n_fields; i++) {
-
+	for (ulint i = 0; i < n_fields; i++) {
 		const byte*	field;
 		ulint		len;
-		ulint		field_no;
 		ulint		orig_len;
-		bool		is_virtual;
 
 		upd_field = upd_get_nth_field(update, i);
-		field_no = mach_read_next_compressed(&ptr);
+		ulint field_no = mach_read_next_compressed(&ptr);
 
-		is_virtual = (field_no >= REC_MAX_N_FIELDS);
+		const bool is_virtual = (field_no >= REC_MAX_N_FIELDS);
 
 		if (is_virtual) {
 			/* If new version, we need to check index list to figure
@@ -1567,15 +1628,63 @@ trx_undo_update_rec_get_update(
 			}
 
 			upd_field_set_v_field_no(upd_field, field_no, index);
+		} else if (UNIV_UNLIKELY((update->info_bits
+					  & ~REC_INFO_DELETED_FLAG)
+					 == REC_INFO_MIN_REC_FLAG)) {
+			ut_ad(type == TRX_UNDO_UPD_EXIST_REC);
+			const ulint uf = index->first_user_field();
+			ut_ad(field_no >= uf);
+
+			if (update->info_bits != REC_INFO_MIN_REC_FLAG) {
+				/* Generic instant ALTER TABLE */
+				if (field_no == uf) {
+					upd_field->new_val.type
+						.metadata_blob_init();
+				} else if (field_no >= index->n_fields) {
+					/* This is reachable during
+					purge if the table was emptied
+					and converted to the canonical
+					format on a later ALTER TABLE.
+					In this case,
+					row_purge_upd_exist_or_extern()
+					would only be interested in
+					freeing any BLOBs that were
+					updated, that is, the metadata
+					BLOB above.  Other BLOBs in
+					the metadata record are never
+					updated; they are for the
+					initial DEFAULT values of the
+					instantly added columns, and
+					they will never change.
+
+					Note: if the table becomes
+					empty during ROLLBACK or is
+					empty during subsequent ALTER
+					TABLE, and btr_page_empty() is
+					called to re-create the root
+					page without the metadata
+					record, in that case we should
+					only free the latest version
+					of BLOBs in the record,
+					which purge would never touch. */
+					field_no = REC_MAX_N_FIELDS;
+					n_skip_field++;
+				} else {
+					dict_col_copy_type(
+						dict_index_get_nth_col(
+							index, field_no - 1),
+						&upd_field->new_val.type);
+				}
+			} else {
+				/* Instant ADD COLUMN...LAST */
+				dict_col_copy_type(
+					dict_index_get_nth_col(index,
+							       field_no),
+					&upd_field->new_val.type);
+			}
+			upd_field->field_no = field_no;
 		} else if (field_no < index->n_fields) {
 			upd_field_set_field_no(upd_field, field_no, index);
-		} else if (update->info_bits == REC_INFO_MIN_REC_FLAG
-			   && index->is_instant()) {
-			/* This must be a rollback of a subsequent
-			instant ADD COLUMN operation. This will be
-			detected and handled by btr_cur_trim(). */
-			upd_field->field_no = field_no;
-			upd_field->orig_len = 0;
 		} else {
 			ib::error() << "Trying to access update undo rec"
 				" field " << field_no
@@ -1608,6 +1717,12 @@ trx_undo_update_rec_get_update(
 			dfield_set_ext(&upd_field->new_val);
 		}
 
+		ut_ad(update->info_bits != (REC_INFO_DELETED_FLAG
+					    | REC_INFO_MIN_REC_FLAG)
+		      || field_no != index->first_user_field()
+		      || (upd_field->new_val.ext
+			  && upd_field->new_val.len == FIELD_REF_SIZE));
+
 		if (is_virtual) {
 			upd_field->old_v_val = static_cast<dfield_t*>(
 				mem_heap_alloc(
@@ -1625,31 +1740,23 @@ trx_undo_update_rec_get_update(
 		}
 	}
 
-	/* In rare scenario, we could have skipped virtual column (as they
-	are dropped. We will regenerate a update vector and skip them */
-	if (n_skip_field > 0) {
-		ulint	n = 0;
-		ut_ad(n_skip_field <= n_fields);
+	/* We may have to skip dropped indexed virtual columns.
+	Also, we may have to trim the update vector of a metadata record
+	if dict_index_t::clear_instant_alter() was invoked on the table
+	later, and the number of fields no longer matches. */
 
-		upd_t*	new_update = upd_create(
-			n_fields + 2 - n_skip_field, heap);
+	if (n_skip_field) {
+		upd_field_t* d = upd_get_nth_field(update, 0);
+		const upd_field_t* const end = d + n_fields + 2;
 
-		for (i = 0; i < n_fields + 2; i++) {
-			upd_field = upd_get_nth_field(update, i);
-
-			if (upd_field->field_no == REC_MAX_N_FIELDS) {
-				continue;
+		for (const upd_field_t* s = d; s != end; s++) {
+			if (s->field_no != REC_MAX_N_FIELDS) {
+				*d++ = *s;
 			}
-
-			upd_field_t*	new_upd_field
-				 = upd_get_nth_field(new_update, n);
-			*new_upd_field = *upd_field;
-			n++;
 		}
-		ut_ad(n == n_fields + 2 - n_skip_field);
-		*upd = new_update;
-	} else {
-		*upd = update;
+
+		ut_ad(d + n_skip_field == end);
+		update->n_fields = d - upd_get_nth_field(update, 0);
 	}
 
 	return(const_cast<byte*>(ptr));
@@ -1704,8 +1811,11 @@ trx_undo_rec_get_partial_row(
 		if (uf->old_v_val) {
 			continue;
 		}
-		ulint c = dict_index_get_nth_col(index, uf->field_no)->ind;
-		*dtuple_get_nth_field(*row, c) = uf->new_val;
+		const dict_col_t& c = *dict_index_get_nth_col(index,
+							      uf->field_no);
+		if (!c.is_dropped()) {
+			*dtuple_get_nth_field(*row, c.ind) = uf->new_val;
+		}
 	}
 
 	end_ptr = ptr + mach_read_from_2(ptr);
@@ -1716,7 +1826,6 @@ trx_undo_rec_get_partial_row(
 		const byte*	field;
 		ulint		field_no;
 		const dict_col_t* col;
-		ulint		col_no;
 		ulint		len;
 		ulint		orig_len;
 		bool		is_virtual;
@@ -1744,15 +1853,18 @@ trx_undo_rec_get_partial_row(
 			dict_v_col_t* vcol = dict_table_get_nth_v_col(
 						index->table, field_no);
 			col = &vcol->m_col;
-			col_no = dict_col_get_no(col);
 			dfield = dtuple_get_nth_v_field(*row, vcol->v_pos);
 			dict_col_copy_type(
 				&vcol->m_col,
 				dfield_get_type(dfield));
 		} else {
 			col = dict_index_get_nth_col(index, field_no);
-			col_no = dict_col_get_no(col);
-			dfield = dtuple_get_nth_field(*row, col_no);
+
+			if (col->is_dropped()) {
+				continue;
+			}
+
+			dfield = dtuple_get_nth_field(*row, col->ind);
 			ut_ad(dfield->type.mtype == DATA_MISSING
 			      || dict_col_type_assert_equal(col,
 							    &dfield->type));
@@ -1760,9 +1872,7 @@ trx_undo_rec_get_partial_row(
 			      || dfield->len == len
 			      || (len != UNIV_SQL_NULL
 				  && len >= UNIV_EXTERN_STORAGE_FIELD));
-			dict_col_copy_type(
-				dict_table_get_nth_col(index->table, col_no),
-				dfield_get_type(dfield));
+			dict_col_copy_type(col, dfield_get_type(dfield));
 		}
 
 		dfield_set_data(dfield, field, len);
@@ -2385,7 +2495,8 @@ trx_undo_prev_version_build(
 		}
 
 		/* Get number of externally stored columns in updated record */
-		const ulint n_ext = dtuple_get_n_ext(entry);
+		const ulint n_ext = index->is_primary()
+			? dtuple_get_n_ext(entry) : 0;
 
 		buf = static_cast<byte*>(mem_heap_alloc(
 			heap, rec_get_converted_size(index, entry, n_ext)));
@@ -2412,7 +2523,8 @@ trx_undo_prev_version_build(
 	rec_offs offsets_dbg[REC_OFFS_NORMAL_SIZE];
 	rec_offs_init(offsets_dbg);
 	ut_a(!rec_offs_any_null_extern(
-		*old_vers, rec_get_offsets(*old_vers, index, offsets_dbg, true,
+		*old_vers, rec_get_offsets(*old_vers, index, offsets_dbg,
+					   index->n_core_fields,
 					   ULINT_UNDEFINED, &heap)));
 #endif // defined UNIV_DEBUG || defined UNIV_BLOB_LIGHT_DEBUG
 
